package tools.descartes.dql.connector.kieker.configgenerator;

import java.io.File;
import java.io.IOException;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Attr;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;

public class XMLGenerator {

	private static final String XML_ELEMENT_ASPECTJ = "aspectj";
	private static final String XML_ELEMENT_WEAVER = "weaver";
	private static final String XML_ATTRIBUTE_OPTIONS = "options";
	private static final String XML_ELEMENT_INCLUDE = "include";
	private static final String XML_ELEMENT_EXCLUDE = "exclude";
	private static final String XML_ATTRIBUTE_WITHIN = "within";
	private static final String XML_ELEMENT_ASPECTS = "aspects";
	private static final String XML_ELEMENT_ASPECT = "aspect";
	private static final String XML_ATTRIBUTE_NAME = "name";
	
	private static final String XML_DECLARATION = "<!DOCTYPE aspectj PUBLIC \"-//AspectJ//DTD//EN\" \"http://www.aspectj.org/dtd/aspectj_1_5_0.dtd\">";

	private static DocumentBuilderFactory docFactory;
	private static DocumentBuilder docBuilder;
	private static Document doc;
	private static TransformerFactory transformerFactory;
	private static Transformer transformer;

	private static void init() throws ParserConfigurationException, TransformerConfigurationException {
		// Documentbuilder declarations
		docFactory = DocumentBuilderFactory.newInstance();
		docBuilder = docFactory.newDocumentBuilder();

		// Document element
		doc = docBuilder.newDocument();
		// declare transformers
		transformerFactory = TransformerFactory.newInstance();
		transformer = transformerFactory.newTransformer();
		//Remove xml declaration in first line
		transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
		//Human readability
		transformer.setOutputProperty(OutputKeys.INDENT, "yes");
		transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
	}

	public static void createXml(String path, String weaverOptions, List<String> includeOptions,
			List<String> excludeOptions, List<String> aspectNames)
			throws ParserConfigurationException, TransformerException, IOException {

		init();
		
		//Doctype Declaration
		DOMImplementation domImpl = doc.getImplementation();
		DocumentType doctype = domImpl.createDocumentType("aspectj", "-//AspectJ//DTD//EN", "http://www.aspectj.org/dtd/aspectj_1_5_0.dtd");
		transformer.setOutputProperty(OutputKeys.DOCTYPE_PUBLIC, doctype.getPublicId());
		transformer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM, doctype.getSystemId());
		doc.appendChild(doctype);

		// <aspectj>
		Element aspectjElement = appendElement(XML_ELEMENT_ASPECTJ);
		// <weaver options="">
		Element weaverElement = appendElement(aspectjElement, XML_ELEMENT_WEAVER, XML_ATTRIBUTE_OPTIONS, weaverOptions);
		// <include withhin="">
		for (String includeOption : includeOptions) {

			// includeoption can be class name or package name
			// We need to include wildcards for all cases
			// Wildcard explained:
			// * - Name continues
			// *. - previous element must exist and can be named anything
			// *.. - previous elements (one or many) must exist and can be named
			// anything
			// .* - next element must exist and can be named anything
			// ..* - next elements (one or many) must exist and can be name
			// anything
			//
			// For some reason Kieker does not monitor when you use a *..
			// wildcard but there is no previous element
			// so all cases must be accounted for.

			String insertOption = "*..*" + includeOption + "*..*";
			appendElement(weaverElement, XML_ELEMENT_INCLUDE, XML_ATTRIBUTE_WITHIN, insertOption);

			insertOption = "*" + includeOption + "*..*";
			appendElement(weaverElement, XML_ELEMENT_INCLUDE, XML_ATTRIBUTE_WITHIN, insertOption);

			insertOption = "*..*" + includeOption + "*";
			appendElement(weaverElement, XML_ELEMENT_INCLUDE, XML_ATTRIBUTE_WITHIN, insertOption);

			insertOption = "*" + includeOption + "*";
			appendElement(weaverElement, XML_ELEMENT_INCLUDE, XML_ATTRIBUTE_WITHIN, insertOption);
		}
		// <exclude withhin="">
		for (String excludeOption : excludeOptions) {
			Element includeElement = appendElement(weaverElement, XML_ELEMENT_EXCLUDE, XML_ATTRIBUTE_WITHIN,
					excludeOption);
		}

		// <aspects>
		Element aspectsElement = appendElement(aspectjElement, XML_ELEMENT_ASPECTS, "", "");

		// <aspect name ="">
		for (String aspectName : aspectNames) {
			Element aspectElement = appendElement(aspectsElement, XML_ELEMENT_ASPECT, XML_ATTRIBUTE_NAME, aspectName);
		}

		DOMSource source = new DOMSource(doc);
		StreamResult result = createFile(path);

		transformer.transform(source, result);

	}

	private static Element appendElement(Element parent, String elementName, String attributeName,
			String attributeValue) {
		Element element = createElement(elementName, attributeName, attributeValue);
		parent.appendChild(element);
		return element;
	}

	private static Element appendElement(String elementName) {
		Element element = createElement(elementName, "", "");
		doc.appendChild(element);
		return element;
	}

	private static Element createElement(String elementName, String attributeName, String attributeValue) {
		System.out.println("Adding:"+elementName+" "+attributeName);
		Element element = doc.createElement(elementName);
		if (!attributeName.equals("")) {
			Attr attr = createAttribute(attributeName);
			attr.setValue(attributeValue);
			element.setAttributeNode(attr);
		}
		return element;
	}

	private static Attr createAttribute(String attributeName) {

		Attr attr = doc.createAttribute(attributeName);
		return attr;

	}

	private static StreamResult createFile(String path) throws IOException {
		
		File f = new File(path);
		if(!f.exists())
		    f.createNewFile();
		System.out.println("File:" + f.getAbsolutePath());
		StreamResult result = new StreamResult(f);
		return result;
	}

}

